package pt.moth.arthoApproach;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import pt.moth.processAnalysis.Process;
import pt.moth.util.IAccess;
import pt.moth.util.SpecialAccess;
import pt.moth.util.View;
import pt.moth.util.variables.Variable;
import pt.moth.viewsAnalysis.Lattice;
import pt.moth.viewsAnalysis.ResultsCollector;
import soot.SootMethod;

/**
 * This class is similar to "ViewCollector" but without the read and write view separation.
 * This will be used by the ArthoSensor
 *  
 * @author Vasco Pessanha
 *
 */
public class ArthoViewsCollector {

	// List of all the processes
	private static Set<Process> processes = new HashSet<Process>();
	// For each Process this attribute keeps its read views
	private static Map<String, List<View>> views = new HashMap<String, List<View>>();
	
	//variables to exclude/not consider in the view creation
	private static final String[] VARIABLES_TO_EXCLUDE = {
			"java.util.concurrent.atomic.AtomicLong",
			"java.lang.Throwable"};
		
	/**
	 * Add Process (with its Views)...
	 * @param p process being added
	 */
	public static void putProcess(Process p) {
		if(!ResultsCollector.hasResults()){
			System.err.println("ViewsCollector - ERROR! Try to put process before compute views analysis");
			return;
		}
		
		List<View> list = new ArrayList<View>();
		
		Set<SootMethod> methods = p.getMethods();
		for(SootMethod m : methods){
			Lattice lat = ResultsCollector.getLattice(m.getSignature());
			View v = extractView(m, lat);
			if(v != null)
				list.add(v);
		}
		if (list.size() > 0) {
			processes.add(new Process(p));
			views.put(p.getName(), list);
		}
	}
	private static View extractView(SootMethod method, Lattice lat) {
		Lattice write_lattice = new Lattice();
		for (IAccess a : lat.toList()) {
			if (!(a instanceof SpecialAccess) && !toExclude(a.getVariable())) {
				write_lattice.add(a);
			}
		}
		if (write_lattice.size() <= 0) {
			return null;
		}
		return new View(write_lattice, method);
	}

	/**
	 * Determines whether or not a given variable should be excluded
	 * @return true - should be excluded, false - should not be excluded
	 */
	private static boolean toExclude(Variable variable) {
		for(String s : VARIABLES_TO_EXCLUDE){
			if(variable == null || variable.toString().contains(s)){
				return true;
			}
		}
		return false;
	}
	/**
	 * Return all views needed by each sensor
	 * @return set of entries of type (Process -> Views)
	 */
	public static Set<Entry<String, List<View>>> getViews(){
		return views.entrySet();
	}
	
	/**
	 * Return all processes
	 * @return all computed processes
	 */
	public static Set<Process> getProcesses(){
		return processes;
	}
	// Printing all methodes' views
	public static String printMethodsViews() {
		String result = "@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				+ "@@  Print Methods Views  @@\n"
				+ "@@@@@@@@@@@@@@@@@@@@@@@@@@@\n\nViews:\n";

		for (Entry<String, List<View>> e : views.entrySet()) {
			result +=e.getKey()+" => "+e.getValue() + "\n";
		}
		result +="\n\nWrite Views:\n";
		return result;
	}
}
