package pt.moth.dependencies;

import java.util.List;
import java.util.Map;
import java.util.Set;

import pt.moth.dependencies.firsPhase.GraphLattice;
import pt.moth.util.Pair;
import pt.moth.util.graphs.DepEdge;
import pt.moth.util.variables.LocalVar;
import pt.moth.util.variables.ReturnRef;
import pt.moth.util.variables.Variable;


/**
 * This class is used to store the graph will all dependencies of all methods of
 * the program. This is passed from the first to second phase of the
 * dependencies process. Some control dependencies are also added in the second
 * phase
 * 
 * @author Vasco Pessanha
 * 
 */
public class VarDepCollector {
	
	// Graph with all dependencies between variables
	protected static VariableGraph complete_graph = new VariableGraph();
	// Graph that only have the dependencies between important variables
	protected static VariableGraph prunned_graph = new VariableGraph();
	protected static final boolean PRINT_PRUNNING = true;

	public static void storeDataDeps(GraphLattice depLat) {
		complete_graph = depLat.getGraph();
	}

	public static VariableGraph getGraph() {
		return complete_graph;
	}
	
	public static VariableGraph getPrunnedGraph() {
		return prunned_graph;
	}
	public static void addEdge(DepEdge e) {
		complete_graph.addEdge(e);
	}

	public static void printDepGraph() {
		complete_graph.printGraph();
	}

	/**
	 * Prune the graph so it only has real/important dependencies. For every
	 * important vertice, flood the graph to see which important vertices are
	 * reached from the first.
	 * 
	 * Another algorithm is described in "pruneGraph2()"'s method
	 */
	public static void pruneGraph() {
		prunned_graph = new VariableGraph();
		Map<VarVersion, Boolean> map = complete_graph.getImportantVertices();
		
		// Statistics
		if (PRINT_PRUNNING){
			int total = 0;
			int importants = 0;
			int jdks = 0;
			int couldBePrunned = 0; //Jdk and Locals
			for(VarVersion g : complete_graph.getVertices()){
				if(map.containsKey(g)){
					importants++;
				}
				else if(isJDKmethod(g)){
					couldBePrunned += (g.getVariable() instanceof LocalVar)?1:0;
					jdks++;
				}
				total++;
			}
			System.out.println("Vertices:\nTotal = "+total+" Jdks = "+jdks+"("+couldBePrunned+") - "+(couldBePrunned*100/total)+"%\n");
			Set<DepEdge> edges = complete_graph.edges();
			int prunnedEdges = 0;
			for(DepEdge e : edges){
				VarVersion origin = e.getOrigin();
				VarVersion dest = e.getDest();
				if(canBePrunned(origin,map) || canBePrunned(dest,map)){
					prunnedEdges++;
				}
			}
			System.out.println("Edges:\nTotal = "+edges.size()+"("+prunnedEdges+" can be prunned) - "+prunnedEdges*100/edges.size()+"%\n");
			System.out.println("PRUNE: Size = "+importants);
		}
		
		// Statistics
		int count = 1;
		for (VarVersion node : map.keySet()) {
			Variable v = node.getVariable();
			if (PRINT_PRUNNING)
				System.out.println("PRUNE: NODE " + count + " = " + node);
			if(!(v instanceof ReturnRef)){
				List<Pair<VarVersion,Boolean>> importantReached = complete_graph.getImportantReached(node, map);
				for (Pair<VarVersion, Boolean> dest : importantReached) {
					prunned_graph.addEdge(new DepEdge(node, dest.getFirst(), dest.getSecond()));
				}
				count++;
			}
		}
	}
	
	private static boolean canBePrunned(VarVersion g, Map<VarVersion, Boolean> map) {
		if(map.containsKey(g))
			return false;
		if(isJDKmethod(g))
			return  g.getVariable() instanceof LocalVar;
		return false;
	}

	public static boolean isJDKmethod(VarVersion g){
		String m = g.getVariable().toString();
		if(m.startsWith("java."))
			return true;
		if(m.startsWith("javax."))
			return true;
		if(m.startsWith("sun."))
			return true;
		return false;
	}
}