package pt.moth.dependencies.firsPhase;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import pt.moth.dependencies.VarVersion;
import pt.moth.dependencies.VariableGraph;
import pt.moth.util.graphs.DepEdge;
import pt.moth.util.variables.IGlobalVar;
import pt.moth.util.variables.IfConditionVar;
import pt.moth.util.variables.ParamRef;
import pt.moth.util.variables.ParamVar;
import pt.moth.util.variables.ReturnRef;
import pt.moth.util.variables.ReturnVar;
import pt.moth.util.variables.Variable;

import soot.SootMethod;
import soot.toolkits.scalar.AbstractFlowSet;
import soot.toolkits.scalar.FlowSet;

/**
 * This FlowSet/Lattice keeps a graph of variables. Each edge of one variable to
 * another represents a data dependency between them. This allows us to
 * represent all data dependencies of a program
 * 
 * @author Vasco Pessanha
 * 
 */
public class GraphLattice extends AbstractFlowSet {
	
	// Could be another prime number. This number says that a given variable is not defined yet
	public static final int SPECIAL_HASH = 7793;
	
	// Graph with all the dependencies
	protected VariableGraph graph = new VariableGraph();
	// For each variable "v", we keep the possible statements where "v" was defined 
	protected Map<Variable,Set<Integer>> hashCodes = new HashMap<Variable, Set<Integer>>();
	// If we are in a non-atomic method called inside a atomic method "m", this field keeps the value of "m" 
	protected SootMethod atomicMethod = null;
	
	public void setAtomicMethod(SootMethod m) {
		atomicMethod = m;
	}
	public SootMethod getAtomicMethod() {
		return atomicMethod;
	}
	public boolean hasAtomicMethod() {
		return atomicMethod != null;
	}

	public void add(Object param) {
		if(param instanceof DepEdge){
			DepEdge edge = (DepEdge) param;
			graph.addEdge(edge);
		}
		else{
			System.err.println("(DepLattice) ERROR! - Shouldn't happen PARAM = "+param);
		}
	}
	
	/**
	 * Set the hashCode of a given variable
	 * @param graph node that has the variable defined and that statement's hashCode.
	 * 	This value overwrites any other.
	 */
	public void setHashCode(VarVersion node){
		Set<Integer> set = new HashSet<Integer>();
		set.add(node.getHashCode());
		hashCodes.put(node.getVariable(), set);
	}
	
	
	public GraphLattice clone() {
		GraphLattice lat = new GraphLattice();
		lat.setAtomicMethod(atomicMethod);
		lat.mergeHashCodes(hashCodes);
		
		for(DepEdge e : graph.edges()){
			lat.add(e.copy());
		}
		return lat;
	}
	public void setGlobalHashCodes(GraphLattice lat, SootMethod method) {
		hashCodes.clear();
		for(Entry<Variable, Set<Integer>> e : lat.hashCodes.entrySet()){
			if(e.getKey() instanceof IGlobalVar && method.equals(e.getKey().getSootMethod())){
				hashCodes.put(e.getKey(), e.getValue());
			}
		}
	}
	private void mergeHashCodes(Map<Variable, Set<Integer>> param) {
		for(Entry<Variable, Set<Integer>> e : param.entrySet()){
			Set<Integer> set = hashCodes.get(e.getKey());
			if(set == null){
				set = new HashSet<Integer>();
			}
			else{
				set = cloneSet(set);
			}
			for(Integer i : e.getValue()){
				set.add(i);	
			}
			this.hashCodes.put(e.getKey(), set);
		}
	}
	
	public boolean contains(Object param) {
		if(param instanceof DepEdge){
			return graph.containsEdge((DepEdge)param);
		}
		return false;
	}

	public boolean isEmpty() {
		return graph.isEmpty();
	}

	public int size() {
		return graph.size();
	}
	public List<DepEdge> toList() {
		List<DepEdge> res = new ArrayList<DepEdge>();
		for(DepEdge e : graph.edges()){
			res.add(e);
		}
		return res;
	}

	@Override
	public void clear() {
		graph.clear();
		atomicMethod = null;
		hashCodes = new HashMap<Variable, Set<Integer>>();
	}

	@Override
	public void copy(FlowSet out) {
		GraphLattice dest = (GraphLattice)out;
		dest.clear();
		dest.setAtomicMethod(atomicMethod);
		dest.mergeHashCodes(hashCodes);
		for(DepEdge e: graph.edges()){
			dest.add(e.copy());	
		}
	}

	public void union(FlowSet set){
		GraphLattice lat = (GraphLattice) set;
		if(atomicMethod == null){
			atomicMethod = lat.atomicMethod;
			mergeHashCodes(lat.hashCodes);
		}
		for(DepEdge e : lat.toList()){
			add(e);
		}
	}
	@Override
	public void union(FlowSet set1, FlowSet out) {
		GraphLattice dest = (GraphLattice)out;
		GraphLattice aux = (GraphLattice)set1;
		out.clear();
		dest.setAtomicMethod(atomicMethod);
		dest.mergeHashCodes(hashCodes);
		dest.mergeHashCodes(aux.hashCodes);
		for(DepEdge e : graph.edges()){
			out.add(e);
		}
		for(DepEdge e : ((GraphLattice)set1).toList()){
			out.add(e);
		}
	}
	
	@Override
	public Object emptySet() {
		return new GraphLattice();
	}

	public String toString(){
		if(atomicMethod == null)
			return graph.toString();
		return graph.toString()+" ("+atomicMethod.getName()+")";
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result
				+ ((atomicMethod == null) ? 0 : atomicMethod.hashCode());
		result = prime * result + ((graph == null) ? 0 : graph.hashCode());
		result = prime * result
				+ ((hashCodes == null) ? 0 : hashCodes.hashCode());
		return result;
	}


	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		GraphLattice other = (GraphLattice) obj;
		if (atomicMethod == null) {
			if (other.atomicMethod != null)
				return false;
		} else if (!atomicMethod.equals(other.atomicMethod))
			return false;
		if (graph == null) {
			if (other.graph != null)
				return false;
		} else if (!graph.equals(other.graph))
			return false;
		if (hashCodes == null) {
			if (other.hashCodes != null)
				return false;
		} else if (!hashCodes.equals(other.hashCodes))
			return false;
		return true;
	}
	@Override
	public Iterator<DepEdge> iterator() {
		return graph.iterator();
	}

	// public boolean containsVertex(GraphNode var) {
	// return graph.containsVertex(var);
	// }

	public VariableGraph getGraph() {
		return graph;
	}
	
	
	// Not implemented Methods:
	
	
	@Override
	public void add(Object paramObject, FlowSet paramFlowSet) {
		System.err.println("(DepLattice)ERROR! - method add(Object paramObject, FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet) {
		System.err.println("(DepLattice)ERROR! - method difference(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err.println("(DepLattice)ERROR! - method difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert(false);
	}
	@Override
	public void intersection(FlowSet paramFlowSet) {
		System.err.println("(DepLattice)ERROR! - method intersection(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err.println("(DepLattice)ERROR! - method intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert(false);
	}

	@Override
	public void remove(Object obj, FlowSet flow) {
		System.err.println("(DepLattice)ERROR! - method remove(Object obj, FlowSet flow)");
		assert(false);
	}
	@Override
	public void remove(Object arg) {
		System.err.println("(DepLattice)ERROR! - method remove(Object arg)");
		assert(false);
	}
	public Set<Integer> getHashCodes(Variable v1) {
		// This type of vars have hashCode = 0
		if(v1 instanceof ReturnVar || v1 instanceof ReturnRef ||
				v1 instanceof ParamVar || v1 instanceof ParamRef || 
				v1 instanceof IfConditionVar){
			Set<Integer> set = new HashSet<Integer>();
			set.add(0);
			return set;
		}
		for(Entry<Variable, Set<Integer>> e : hashCodes.entrySet()){
			if(e.getKey().equals(v1)){
				return cloneSet(e.getValue());
			}
		}
		Set<Integer> set = new HashSet<Integer>();
		set.add(SPECIAL_HASH);
		return set;
	}
	private Set<Integer> cloneSet(Set<Integer> value) {
		Set<Integer> set = new HashSet<Integer>();
		for(Integer i : value){
			set.add(new Integer(i));
		}
		return set;
	}
}
