package pt.moth.instanceTypeAnalysis.firstPhase;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import pt.moth.util.Pair;
import pt.moth.util.variables.LocalVar;


import soot.toolkits.scalar.AbstractFlowSet;
import soot.toolkits.scalar.FlowSet;

/**
 * This FlowSet/Lattice allows us to keep the information of an implementing class of a local variable.
 * If we have "entries = new HashMap()" this will be transformed in the following JimpleCode:
 * $r0 = new HashMap();
 * entries = $r0
 * 
 * Everytime we have a method call we pass the empty lattice 
 * @author Vasco Pessanha
 * 
 */
public class ClassLattice extends AbstractFlowSet {

	protected Set<Pair<LocalVar, String>> variables = new HashSet<Pair<LocalVar,String>>();

	//param has type Pair<LocalVar, Set<String>>
	public void add(Object param) {
		if(param instanceof Pair<?,?> && 
			((Pair<?,?>)param).getFirst() instanceof LocalVar &&
				((Pair<?,?>)param).getSecond() instanceof String){
			LocalVar first = (LocalVar)((Pair<?,?>)param).getFirst();
			String second = (String)((Pair<?,?>)param).getSecond();
			variables.add(new Pair<LocalVar, String>(first,second));
		}
		else{
			System.err.println("(Lattice) ERROR! - Shouldn't happen PARAM = "+param);
		}
	}
	public AbstractFlowSet clone() {
		ClassLattice abs = new ClassLattice();
		for(Pair<LocalVar, String> p : variables){
			abs.variables.add(new Pair<LocalVar, String>(p));
		}
		return abs;
	}


	//param has type Pair<LocalVar, Set<String>>
	public boolean contains(Object param) {
		if(param instanceof Pair<?,?> &&
			((Pair<?,?>)param).getFirst() instanceof LocalVar && 
				((Pair<?,?>)param).getSecond() instanceof String){
			LocalVar first = (LocalVar)((Pair<?,?>)param).getFirst();
			String second = (String)((Pair<?,?>)param).getSecond();
			return variables.contains(new Pair<LocalVar, String>(first,second));
		}
		else{
			System.err.println("(Lattice) ERROR! - CONTAINS, PARAM = "+param);
		}
		return false;
	}

	public boolean isEmpty() {
		return variables.isEmpty();
	}

	//param has type LocalVar
	public void remove(Object param) {
		if(param instanceof Pair<?,?> && 
			((Pair<?,?>)param).getFirst() instanceof LocalVar &&
				((Pair<?,?>)param).getSecond() instanceof String){
			LocalVar first = (LocalVar)((Pair<?,?>)param).getFirst();
			String second = (String)((Pair<?,?>)param).getSecond();
			Pair<LocalVar,String> p = new Pair<LocalVar, String>(first,second);
			if(variables.contains(p)){
				variables.remove(new Pair<LocalVar, String>(first,second));
			}
		}
		else{
			System.err.println("(Lattice) ERROR! - REMOVE, PARAM = "+param);
		}
	}

	public int size() {
		return variables.size();
	}
	public List<Pair<LocalVar,String>> toList() {
		List<Pair<LocalVar,String>> list = new ArrayList<Pair<LocalVar,String>>();
		for(Pair<LocalVar, String> pair : variables){
			list.add(new Pair<LocalVar,String>(pair));
		}
		return list;
	}

	@Override
	public void clear() {
		variables.clear();
	}

	@Override
	public void copy(FlowSet dest) {
		dest.clear();
		for(Pair<LocalVar, String> p : variables){
			((ClassLattice)dest).add(new Pair<LocalVar, String>(p));	
		}
	}

	public void union(FlowSet set){
		ClassLattice lat = (ClassLattice) set;
		for(Pair<LocalVar, String> p : lat.toList()){
			add(new Pair<LocalVar,String>(p));
		}
	}
	@Override
	public void union(FlowSet set1, FlowSet out) {
		out.clear();
		for(Pair<LocalVar, String> p : ((ClassLattice)set1).toList()){
			out.add(new Pair<LocalVar,String>(p));
		}
		for(Pair<LocalVar, String> p : variables){
			out.add(new Pair<LocalVar,String>(p));
		}
	}
	
	@Override
	public Object emptySet() {
		return new ClassLattice();
	}

	public Set<String> getImplementations(LocalVar ref) {
		Set<String> results = new HashSet<String>();
		for(Pair<LocalVar, String> p : variables){
			if (p.getFirst().equals(ref)){
				results.add(p.getSecond());
			}
		}
		return results;
	}
	
	public void addParamClasses(LocalVar old_var, LocalVar new_var){
		Set<Pair<LocalVar,String>> toAdd = new HashSet<Pair<LocalVar,String>>();
		for(Pair<LocalVar, String> pair :variables){
			if(pair.getFirst().equals(old_var)){
				toAdd.add(new Pair<LocalVar,String>(new_var,pair.getSecond()));
			}
		}
		for(Pair<LocalVar,String> pair : toAdd){
			variables.add(pair);	
		}
	}
	public void removeParamClasses(LocalVar old){
		Set<Pair<LocalVar,String>> pairsToRemove = new HashSet<Pair<LocalVar,String>>();
		for(Pair<LocalVar, String> p : variables){
			if(p.getFirst().equals(old)){
				pairsToRemove.add(p);
			}
		}
		for(Pair<LocalVar, String> p : pairsToRemove)
			remove(p);
	}
	
	
	public String toString(){
		if(isEmpty())
			return "{}";
		String result = "{";
		for(Pair<LocalVar, String> e : variables){
			result+="("+e.getFirst()+","+e.getSecond()+"),";
		}
		return result.substring(0,result.length()-1)+"}";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result
				+ ((variables == null) ? 0 : variables.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if(!(obj instanceof ClassLattice) || obj == null)
			return false;
		ClassLattice other = (ClassLattice) obj;
		if(variables == null || other.variables == null || other.size() != this.size())
			return false;
		for(Pair<LocalVar, String> p : other.toList()){
			if(!variables.contains(p)){
				return false;
			}
		}
		return true;
	}

	@Override
	public Iterator<Pair<LocalVar,String>> iterator() {
		return toList().iterator();
	}
	@Override
	public void add(Object paramObject, FlowSet paramFlowSet) {
		System.err.println("(Lattice)ERROR! - method add(Object paramObject, FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet) {
		System.err.println("(Lattice)ERROR! - method difference(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err.println("(Lattice)ERROR! - method difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert(false);
	}
	@Override
	public void intersection(FlowSet paramFlowSet) {
		System.err.println("(Lattice)ERROR! - method intersection(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err.println("(Lattice)ERROR! - method intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert(false);
	}

	@Override
	public void remove(Object paramObject, FlowSet paramFlowSet) {
		System.err.println("(Lattice)ERROR! - method remove");
		assert(false);
	}
}
