package pt.moth.instanceTypeAnalysis.secondPhase;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import pt.moth.util.Pair;
import pt.moth.util.variables.GlobalVar;
import pt.moth.util.variables.LocalVar;


import soot.toolkits.scalar.AbstractFlowSet;
import soot.toolkits.scalar.FlowSet;

/**
 * This FlowSet/Lattice just keeps the relation between method parameters and
 * its global reference. For example, if we have the method "add(List l)" and
 * the param l has the global field "ourList" then we keep a pair <l, ourList>
 * 
 * This information is used to detect the possible implementing classes of a
 * global field
 * 
 * @author Vasco Pessanha
 * 
 */
public class ParamLattice extends AbstractFlowSet {

	protected Map<LocalVar, GlobalVar> variables = new HashMap<LocalVar, GlobalVar>();

	// param has type Pair<LocalVar, Variable>
	public void add(Object param) {
		if (param instanceof Pair<?, ?>
				&& ((Pair<?, ?>) param).getFirst() instanceof LocalVar
				&& ((Pair<?, ?>) param).getSecond() instanceof GlobalVar) {
			LocalVar first = (LocalVar) ((Pair<?, ?>) param).getFirst();
			GlobalVar second = (GlobalVar) ((Pair<?, ?>) param).getSecond();
			variables.put(first, second);
		} else {
			System.err.println("(Lattice) ERROR! - Shouldn't happen PARAM = "
					+ param);
		}
	}

	public AbstractFlowSet clone() {
		ParamLattice abs = new ParamLattice();
		for (Entry<LocalVar, GlobalVar> e : variables.entrySet()) {
			abs.variables.put(e.getKey(), e.getValue());
		}
		return abs;
	}

	// param has type LocalVar
	public boolean contains(Object param) {
		if (param instanceof LocalVar) {
			return variables.get((LocalVar) param) != null;
		} else {
			System.err.println("(Lattice) ERROR! - CONTAINS, PARAM = " + param);
			return false;
		}
	}

	public boolean isEmpty() {
		return variables.isEmpty();
	}

	// param has type LocalVar
	public void remove(Object param) {
		if (param instanceof LocalVar) {
			variables.remove((LocalVar) param);
		} else {
			System.err.println("(Lattice) ERROR! - REMOVE, PARAM = " + param);
		}
	}

	public int size() {
		return variables.size();
	}

	public List<Pair<LocalVar, GlobalVar>> toList() {
		List<Pair<LocalVar, GlobalVar>> list = new ArrayList<Pair<LocalVar, GlobalVar>>();
		for (Entry<LocalVar, GlobalVar> e : variables.entrySet()) {
			list.add(new Pair<LocalVar, GlobalVar>(e.getKey(), e.getValue()));
		}
		return list;
	}

	@Override
	public void clear() {
		variables.clear();
	}

	@Override
	public void copy(FlowSet dest) {
		dest.clear();
		for (Entry<LocalVar, GlobalVar> e : variables.entrySet()) {
			((ParamLattice) dest).add(new Pair<LocalVar, GlobalVar>(e.getKey(), e
					.getValue()));
		}
	}

	public void union(FlowSet set) {
		ParamLattice lat = (ParamLattice) set;
		for (Pair<LocalVar, GlobalVar> p : lat.toList()) {
			add(new Pair<LocalVar, GlobalVar>(p));
		}
	}

	@Override
	public void union(FlowSet set1, FlowSet out) {
		out.clear();
		for (Pair<LocalVar, GlobalVar> p : ((ParamLattice) set1).toList()) {
			out.add(new Pair<LocalVar, GlobalVar>(p));
		}
		for (Entry<LocalVar, GlobalVar> e : variables.entrySet()) {
			out.add(new Pair<LocalVar, GlobalVar>(e.getKey(), e.getValue()));
		}
	}

	@Override
	public Object emptySet() {
		return new ParamLattice();
	}

	public GlobalVar getField(LocalVar local) {
		return variables.get(local);
	}

	public String toString() {
		if (isEmpty())
			return "{}";
		String result = "{";
		for (Entry<LocalVar, GlobalVar> e : variables.entrySet()) {
			result += "(" + e.getKey() + "," + e.getValue() + "),";
		}
		return result.substring(0, result.length() - 1) + "}";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result
				+ ((variables == null) ? 0 : variables.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof ParamLattice) || obj == null)
			return false;
		ParamLattice other = (ParamLattice) obj;
		if (variables == null || other.variables == null
				|| other.size() != this.size())
			return false;
		for (Pair<LocalVar, GlobalVar> p : other.toList()) {
			if (!variables.get(p.getFirst()).equals(p.getSecond())) {
				return false;
			}
		}
		return true;
	}

	@Override
	public Iterator<Pair<LocalVar, GlobalVar>> iterator() {
		return toList().iterator();
	}

	@Override
	public void add(Object paramObject, FlowSet paramFlowSet) {
		System.err
				.println("(Lattice)ERROR! - method add(Object paramObject, FlowSet paramFlowSet)");
		assert (false);
	}

	@Override
	public void difference(FlowSet paramFlowSet) {
		System.err
				.println("(Lattice)ERROR! - method difference(FlowSet paramFlowSet)");
		assert (false);
	}

	@Override
	public void difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err
				.println("(Lattice)ERROR! - method difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert (false);
	}

	@Override
	public void intersection(FlowSet paramFlowSet) {
		System.err
				.println("(Lattice)ERROR! - method intersection(FlowSet paramFlowSet)");
		assert (false);
	}

	@Override
	public void intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err
				.println("(Lattice)ERROR! - method intersection(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert (false);
	}

	@Override
	public void remove(Object paramObject, FlowSet paramFlowSet) {
		System.err.println("(Lattice)ERROR! - method remove");
		assert (false);
	}
}
