package pt.moth.viewsAnalysis;


import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import pt.moth.instanceTypeAnalysis.UnitsCollector;
import pt.moth.inter.MethodUtils;
import pt.moth.util.Access;
import pt.moth.util.ArrayRef;
import pt.moth.util.variables.GlobalVar;
import pt.moth.util.variables.LocalVar;
import pt.moth.xmlAnnotations.AnnotationsParser;
import pt.moth.xmlAnnotations.NativesCollector;


import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.SootMethodRef;
import soot.Type;
import soot.Unit;
import soot.Value;
import soot.jimple.Constant;
import soot.jimple.FieldRef;
import soot.jimple.InvokeExpr;
import soot.jimple.MonitorStmt;
import soot.jimple.StaticFieldRef;
import soot.jimple.Stmt;
import soot.jimple.internal.AbstractBinopExpr;
import soot.jimple.internal.AbstractInstanceOfExpr;
import soot.jimple.internal.AbstractLengthExpr;
import soot.jimple.internal.AbstractUnopExpr;
import soot.jimple.internal.JAddExpr;
import soot.jimple.internal.JArrayRef;
import soot.jimple.internal.JAssignStmt;
import soot.jimple.internal.JCastExpr;
import soot.jimple.internal.JGotoStmt;
import soot.jimple.internal.JIdentityStmt;
import soot.jimple.internal.JIfStmt;
import soot.jimple.internal.JInstanceFieldRef;
import soot.jimple.internal.JInterfaceInvokeExpr;
import soot.jimple.internal.JLookupSwitchStmt;
import soot.jimple.internal.JNewArrayExpr;
import soot.jimple.internal.JNewExpr;
import soot.jimple.internal.JNewMultiArrayExpr;
import soot.jimple.internal.JReturnStmt;
import soot.jimple.internal.JReturnVoidStmt;
import soot.jimple.internal.JSpecialInvokeExpr;
import soot.jimple.internal.JStaticInvokeExpr;
import soot.jimple.internal.JTableSwitchStmt;
import soot.jimple.internal.JThrowStmt;
import soot.jimple.internal.JVirtualInvokeExpr;
import soot.jimple.internal.JimpleLocal;
import soot.toolkits.graph.DirectedGraph;
import soot.toolkits.graph.UnitGraph;
import soot.toolkits.scalar.ForwardFlowAnalysis;

public class IntraAnalysisAtomic extends ForwardFlowAnalysis<Unit, Lattice>{

	protected InterAnalysis interAnalysis;
	protected SootMethod thisMethod;
	
	/**
	 * This structure allows us to keep the set of readVariables
	 * that are stored in the outValue Lattice in the Return Stmt
	 */
	protected Map<LocalVar, FieldRef> readVariables;
	protected Map<LocalVar, ArrayRef> arrayVariables;
	
	//Method's annotations generated by the XML File..
	protected AnnotationsParser methodAnnotations;
	
	public IntraAnalysisAtomic(DirectedGraph<Unit> graph, InterAnalysis interAnalysis, AnnotationsParser methodAnnotations) {
		super(graph);
		this.interAnalysis = interAnalysis;
		this.thisMethod = ((UnitGraph) graph).getBody().getMethod();
		this.readVariables = new HashMap<LocalVar, FieldRef>();
		this.arrayVariables = new HashMap<LocalVar, ArrayRef>();
		this.methodAnnotations = methodAnnotations;
		doAnalysis();
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void flowThrough(Lattice inValue, Unit unit, Lattice outValue) {
		inValue.copy(outValue);
		// @@@@@@@@@	Do not analyse exceptions	@@@@@@@@@
		SootClass superClass = thisMethod.getDeclaringClass();
		while(superClass.hasSuperclass() && !superClass.getSuperclass().toString().contains("java.lang.Object")){
			superClass = superClass.getSuperclass();
		}
		if(superClass.toString().contains("Throwable")){
			return;
		}
		// @@@@@@@@		Analyse if not exception    @@@@@@@@@
		if (((Stmt) unit).containsInvokeExpr()) {
			InvokeExpr expr = ((Stmt)unit).getInvokeExpr();
			SootMethod method = expr.getMethod();
			boolean hasAnnotations = methodAnnotations.hasInfo(method);
			JimpleLocal local = getInvokeExprLocal(expr);
			
			//This method has annotations (we dont have to analyse it)
			if(hasAnnotations){
				NativesCollector.usedAnnotations(method);
				// remove redundant read
				FieldRef var = readVariables.remove(new LocalVar(thisMethod,local));
				//This method was called in a importante variable
				if(var != null){
				 	//changesObject?
					if(methodAnnotations.changesObject(method)){
						outValue.add(new Access(true,new GlobalVar(var)));
						excludeRedundantReads(var);
					}
					//readsObject?
					if(methodAnnotations.readsObject(method)){
						outValue.add(new Access(false, new GlobalVar(var)));
					}
				}
				else{}//this method was called in a non important local variable
			}
			//We cannot analyse this method!
			else if(!interAnalysis.shouldAnalyzeMethod(method)){
				Set<String> implementations = UnitsCollector.getImplementations(unit);
				//but we have implementations (e.g. List => ArrayList)
				if(implementations.size() > 0){
					excludeRedundantReads(local);
					Set<String> methodsToStay = new HashSet<String>();
					SootMethodRef ref;
					for(String i : implementations){
						ref = Scene.v().makeMethodRef(new SootClass(i), method.getName(), method.getParameterTypes(), method.getReturnType(), method.isStatic());
						methodsToStay.add(ref.resolve().getSignature());
					}
					MethodUtils.changeEdges(thisMethod, (Stmt)unit, methodsToStay);
					Lattice returnValue = interAnalysis.handleMethodInvocation(thisMethod, (Lattice)Lattice.ATOMIC_LATTICE.clone(),(Stmt) unit);
					//this union is the "mergep" function of PADTAD2011
					inValue.union(returnValue, outValue);
				}
				//It is impossible to analyse (assume writes and reads obj)
				else{
					if(!method.isNative()){
						if(!method.getName().equals("<init>") && local != null && readVariables.get(local) != null){
//							System.err.println("(IntraAnalysisAtomic) Phantom method (but not native) => " + method);
							excludeRedundantReads(local);
						}
					}
					else{//native method
						if(local != null){
							LocalVar localVar = new LocalVar(thisMethod, local);
							FieldRef var = readVariables.get(localVar);
							if(var != null){
								outValue.add(new Access(false,new GlobalVar(var)));
								outValue.add(new Access(true, new GlobalVar(var)));
							}
						}
						
						// Read and Write in the global parameters
						Iterator<Value> args = expr.getArgs().iterator();
						while(args.hasNext()){
							Value arg = args.next();
							FieldRef p_ref = null;
							if(arg instanceof JimpleLocal){
								p_ref = readVariables.get(new LocalVar(thisMethod, (JimpleLocal)arg));
							}
							else if(arg instanceof FieldRef){
								p_ref = (FieldRef)arg;
							}
							if(p_ref != null){
								outValue.add(new Access(false,new GlobalVar(p_ref)));
								outValue.add(new Access(true, new GlobalVar(p_ref)));											
							}
						}
						excludeRedundantReads(local);
						NativesCollector.annotate(method, local != null);
					}
				}
			}
			else{
				excludeRedundantReads(local);
				// handle method invocation: flow parameters into method call,
				// and flow return value from method call
				Lattice returnValue = interAnalysis.handleMethodInvocation(thisMethod, Lattice.ATOMIC_LATTICE,
						(Stmt) unit);
				//this union is the "mergep" function of PADTAD2011
				inValue.union(returnValue, outValue);
			}
		}
		// This instruction is an assignment x = 4 or z = x
		else if (unit instanceof JAssignStmt) {
			Value leftop = ((JAssignStmt) unit).getLeftOp();
			Value rightop = ((JAssignStmt) unit).getRightOp();
			
			// Global variable in the left side "x = ...". We have "Write(x)"
			// Local Variables are JimpleLocal variables so they not fit in this test
			if (leftop instanceof JInstanceFieldRef || leftop instanceof StaticFieldRef) {
				FieldRef var = (FieldRef) leftop;
				outValue.add(new Access(true,new GlobalVar(var)));
			}
			// Global variable in the right side "? = x (op ?)?". We have
			// "Read(x)"
			else if (rightop instanceof JInstanceFieldRef || rightop instanceof StaticFieldRef) {
				if (leftop instanceof JimpleLocal) {
				
					Type type = ((FieldRef)rightop).getType();
					LocalVar local = new LocalVar(thisMethod,(JimpleLocal)leftop);
					FieldRef global = (FieldRef)rightop;
					//If it is an array store it in the map
					if(type != null && type.toString().contains("[]")){
						arrayVariables.put(local, new ArrayRef(global, countBrackets(type.toString())));
					}
					//And if it's a global variable...
					else{
						Stmt stmt = (Stmt)thisMethod.getActiveBody().getUnits().getSuccOf(unit);
						if(stmt == null){
							System.err.println("IntraAnalysisAtomic -  Error! Shouldn't happen: "+thisMethod + " "+unit);
						}
						//is an important read or is just used to make another method call?
						//is important read iff this stmt is not followed by an invoke stmt or that
						//invoke stmt is not made to this local
						boolean importantRead = !stmt.containsInvokeExpr();
						if(!importantRead){
							JimpleLocal expr_jimple = getInvokeExprLocal(stmt.getInvokeExpr());
							LocalVar expr_local = new LocalVar(thisMethod,expr_jimple);
							if(expr_jimple == null || !expr_local.equals(local)){
								importantRead = true;
							}
						}
						if(importantRead){
							//is important (keep it)
							outValue.add(new Access(false, new GlobalVar(global)));
						}
						//is not important (keep to the next unit)
						readVariables.put(local,global);
					}
				}
				else {
					System.err.println("(IntraAnalysisAtomic) What Else???\nleftop = "+leftop+"\nrightop = "+rightop);
				}
			}
			else if(leftop instanceof JArrayRef){
				ArrayRef array = arrayVariables.get(new LocalVar(thisMethod,(JimpleLocal)((JArrayRef)leftop).getBase()));				
				if(array != null){
					FieldRef var = array.getArray();
					outValue.add(new Access(true,new GlobalVar(var)));
				}
				else{}//Non important array
			}
			else if(rightop instanceof JArrayRef){
				ArrayRef array = arrayVariables.get(new LocalVar(thisMethod,(JimpleLocal)((JArrayRef)rightop).getBase()));
				if(array != null){
					int level = array.getLevel();
					if(level == 1){
						outValue.add(new Access(false, new GlobalVar(array.getArray())));
					}
					else{
						arrayVariables.put(new LocalVar(thisMethod,(JimpleLocal)leftop), new ArrayRef(array.getArray(), level -1));
					}
				}
				else{}//Non Important array
			}
			else if (leftop instanceof JimpleLocal){
				if(rightop instanceof JCastExpr){}		//Ignorar casts
				else if(rightop instanceof JNewExpr){}	//Ignora expressoes New()..
				else if(rightop instanceof JAddExpr){}	//Ignora somas..
				else if(rightop instanceof JimpleLocal){}	//Ignorar local = local...
				else if(rightop instanceof Constant){}	//Ignorar constantes
				else if(rightop instanceof JNewArrayExpr){}		//Ignorar new ?[]
				else if(rightop instanceof AbstractUnopExpr){
					if(rightop instanceof AbstractLengthExpr){
						AbstractLengthExpr leng_expr = (AbstractLengthExpr) rightop;
						if(leng_expr.getOp() instanceof JimpleLocal){
							LocalVar local = new LocalVar(thisMethod,(JimpleLocal)leng_expr.getOp());
							ArrayRef ref = arrayVariables.get(local);
							if(ref != null){
								outValue.add(new Access(false, new GlobalVar(ref.getArray())));
							}
						}
					}
					//Ignorar expressoes de um arg (e.g. $r0.length())
				}	
				else if(rightop instanceof AbstractBinopExpr){}	//Ignorar expressoes de 2args (e.g. cmp)
				else if(rightop instanceof AbstractInstanceOfExpr){} //Ignorar expr "instanceof"
				else if(rightop instanceof JNewMultiArrayExpr){}	//Ignorar New de arrys multiplos (tds duma vez)
				else{
					System.err.println("\nOther JAssignStmt Struture:");
					System.err.println("Left = " + leftop.getClass() + "("+ leftop + ")");
					System.err.println("Right = " + rightop.getClass() + "("+ rightop + ")\n");
				}
			}
			// Other JAssignStmt Struture
			else {
				System.err.println("\nOther JAssignStmt Struture:");
				System.err.println("Left = " + leftop.getClass() + "("+ leftop + ")");
				System.err.println("Right = " + rightop.getClass() + "("+ rightop + ")\n");
			}
		}
		else if(unit instanceof JReturnStmt || unit instanceof JReturnVoidStmt){}
		else if(unit instanceof JIdentityStmt){}//This can identify Parameters or classes
		else if(unit instanceof JIfStmt){}//This can identify a if with local vars (e.g. if($r0 == null))
		else if(unit instanceof JThrowStmt){}//Throw exception with localVariable (e.g. throw $r0)
		else if(unit instanceof JGotoStmt){}//JGotoStmt used, for instances, in loops
		else if(unit instanceof JTableSwitchStmt){}//JTableSwitchStmt used
		else if(unit instanceof MonitorStmt){}//Monitor Statement
		else if(unit instanceof JLookupSwitchStmt){}//JLookupSwitchStmt used
		
		// Others..
		else {
			System.err.println("(IntraAnalysisAtomic)("+thisMethod.getName()+")Other class : (in "+unit.getClass()+") "+unit);
		}
//		System.out.println("("+thisMethod.getDeclaringClass()+","+thisMethod.getName() + ") " + unit + "("+inValue+","+outValue+")");
	}

	private Integer countBrackets(String string) {
		 char[] str = string.toCharArray();
		 int num_brackets = 0;
		 for(int i = 0; i < str.length -1; i++){
			 if(str[i] == '[' && str[i+1] == ']'){
				 num_brackets++;
				 i++;
			 }
		 }
		 return num_brackets;
	}
	
	//Returns the local variable (Base) of a InvokeExpr.
	//E.g. $r0.getX() returns "$r0"
	private JimpleLocal getInvokeExprLocal(InvokeExpr expr){
		if(expr instanceof JVirtualInvokeExpr && ((JVirtualInvokeExpr)expr).getBase() instanceof JimpleLocal){
			return (JimpleLocal)((JVirtualInvokeExpr)expr).getBase();
		}
		else if(expr instanceof JInterfaceInvokeExpr && ((JInterfaceInvokeExpr)expr).getBase() instanceof JimpleLocal){
			return (JimpleLocal)((JInterfaceInvokeExpr)expr).getBase();
		}
		else if(expr instanceof JSpecialInvokeExpr || expr instanceof JStaticInvokeExpr){
			//Ignore! There is no local variable. Just a variable with the class reference
		}
		else{
			System.err.println("Other InvokeExpr class: " + (expr.getClass()));
		}
		return null;
	}

	private void excludeRedundantReads(FieldRef var){
		if(var instanceof JInstanceFieldRef && ((JInstanceFieldRef)var).getBase() instanceof JimpleLocal){
			JimpleLocal base = ((JimpleLocal)((JInstanceFieldRef)var).getBase());
			readVariables.remove(new LocalVar(thisMethod,base));
		}
	}
	private void excludeRedundantReads(JimpleLocal var){
		readVariables.remove(new LocalVar(thisMethod,var));
	}
	
	@Override
	protected void copy(Lattice source, Lattice dest) {
		source.copy(dest);
	}

	@Override
	protected Lattice entryInitialFlow() {
		return new Lattice();
	}

	@Override
	protected void merge(Lattice in1, Lattice in2, Lattice out) {
		in1.union(in2, out);
	}

	@Override
	protected Lattice newInitialFlow() {
		return new Lattice();
	}

	public Lattice getFlowAtEnd() {
		Lattice res = newInitialFlow();

		// merge the flowsets of all the method's exit-points
		List<Unit> tails = this.graph.getTails();
		for (Unit unit : tails) {
			Lattice sep1 = getFlowAfter(unit);
			if (sep1 == null) { // this method was not analyzed
				return res;
			}
			Lattice tmp = newInitialFlow();
//			sep1.unionReturn(res, tmp);
			sep1.union(res, tmp);
			res = tmp;
		}
		return res;
	}
}
