package pt.moth.xmlAnnotations;

import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Set;
import java.util.TreeSet;

import java.io.File;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import soot.SootMethod;

/**
 * This class receives a xml file and analyse it, parsing every annotation to
 * the Java structures (methodsW and methodsR). Then, it offers methods that can
 * check if a given method reads/writes in the main object or in a given
 * parameter
 * 
 * @author Vasco Pessanha
 * 
 */
public class AnnotationsParser {

	private Map<MethodName, Changes> methodsW;
	private Map<MethodName, Reads> methodsR;

	public AnnotationsParser(String xmlFile) {
		methodsR = new TreeMap<MethodName, Reads>();
		methodsW = new TreeMap<MethodName, Changes>();
		load(xmlFile);
	}

	private void load(String xml) {
		if (new File(xml).exists())
			try {
				loadXML(xml);
			} catch (Exception e) {
				System.out.println("Error loading xml method db: "
						+ e.getMessage());
			}
	}

	private void loadXML(String file) throws Exception {

		Document xml = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().parse(new File(file));
		NodeList classesList = xml.getElementsByTagName("classes");
		// NodeList childs = xml.getChildNodes();
		loadClasses(classesList.item(0));
	}

	// root load (all classes)
	private void loadClasses(Node classes) {
		NodeList childs = classes.getChildNodes();

		for (int i = 0; i < childs.getLength(); i++) {
			Node node = childs.item(i);
			if (node.getNodeName().equals("class")
					|| node.getNodeName().equals("automatic_class"))
				loadClass(node);
		}
	}

	// Class load
	private void loadClass(Node classNode) {
		String classId = classNode.getAttributes().getNamedItem("id")
				.getNodeValue();
		NodeList childs = classNode.getChildNodes();

		for (int i = 0; i < childs.getLength(); i++) {
			Node node = childs.item(i);
			if (node.getNodeName().equals("method"))
				loadMethod(node, classId);
		}
	}

	// load a method
	private void loadMethod(Node methodNode, String classId) {
		String methodId = methodNode.getAttributes().getNamedItem("id")
				.getNodeValue();
		methodId = methodId.trim();
		Changes changes = new Changes();
		Reads reads = new Reads();
		NodeList childs = methodNode.getChildNodes();

		methodsW.put(new MethodName(classId, trimWhiteSpaces(methodId)),
				changes);
		methodsR.put(new MethodName(classId, trimWhiteSpaces(methodId)), reads);

		for (int i = 0; i < childs.getLength(); i++) {
			Node node = childs.item(i);

			if (node.getNodeName().equals("changes")) {
				String value = node.getFirstChild().getNodeValue();

				if (value.equals("this"))
					changes.setChangesObject();
				else
					changes.add(Integer.parseInt(value));
			} else if (node.getNodeName().equals("reads")) {
				String value = node.getFirstChild().getNodeValue();

				if (value.equals("this"))
					reads.setReadsObject();
				else
					reads.add(Integer.parseInt(value));
			}
		}
	}

	public boolean hasInfo(SootMethod method) {
		MethodName aux = new MethodName(method);
		// Returns if this method has XML anotations
		if (methodsW.containsKey(aux) || methodsR.containsKey(aux)) {
			return true;
		}
		return false;
	}

	public boolean readsObject(SootMethod method) {
		MethodName aux = new MethodName(method);
		return methodsR.get(aux).readsObject();
	}

	public boolean changesObject(SootMethod method) {
		MethodName aux = new MethodName(method);
		return methodsW.get(aux).changesObject();
	}

	public boolean argChanged(SootMethod method, int arg) {
		MethodName aux = new MethodName(method);
		return methodsW.get(aux).argChanged(arg);
	}

	public boolean argRead(SootMethod method, int arg) {
		MethodName aux = new MethodName(method);
		return methodsR.get(aux).argRead(arg);
	}

	public String toString() {
		String result = "";
		for (Entry<MethodName, Changes> e : methodsW.entrySet()) {
			result += e.getKey() + " " + e.getValue().toString() + "\n";
		}
		return result;
	}

	private String trimWhiteSpaces(String x) {
		return x.replaceAll(" ", "");
	}

	/**
	 * This class represents the changes of a given method (params and if
	 * changes the main object)
	 * 
	 * @author Vasco Pessanha
	 * 
	 */
	class Changes {
		private boolean changesObject;
		private Set<Integer> argsChanged;

		public Changes() {
			this.changesObject = false;
			argsChanged = new TreeSet<Integer>();
		}

		public boolean changesObject() {
			return changesObject;
		}

		public void setChangesObject() {
			changesObject = true;
		}

		public void add(int arg) {
			argsChanged.add(arg);
		}

		public boolean argChanged(int arg) {
			return argsChanged.contains(arg);
		}

		public String toString() {
			if (!changesObject && argsChanged.size() <= 0)
				return "changes nothing!";
			else if (changesObject && argsChanged.size() <= 0)
				return "changes this object";
			String result = "changes this object and params";
			for (Integer i : argsChanged)
				result += " " + i + ",";
			return result.substring(0, result.length() - 1);
		}
	}

	/**
	 * This class represents the reads of a given method (params and if reads
	 * the main object)
	 * 
	 * @author Vasco Pessanha
	 * 
	 */
	class Reads {
		private boolean readsObject;
		private Set<Integer> argsRead;

		public Reads() {
			this.readsObject = false;
			argsRead = new TreeSet<Integer>();
		}

		public boolean readsObject() {
			return readsObject;
		}

		public void setReadsObject() {
			readsObject = true;
		}

		public void add(int arg) {
			argsRead.add(arg);
		}

		public boolean argRead(int arg) {
			return argsRead.contains(arg);
		}

		public String toString() {
			if (!readsObject && argsRead.size() <= 0)
				return "reads nothing!";
			else if (readsObject && argsRead.size() <= 0)
				return "reads this object";
			String result = "reads this object and params";
			for (Integer i : argsRead)
				result += " " + i + ",";
			return result.substring(0, result.length() - 1);
		}
	}
}
