package pt.moth.viewsAnalysis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import pt.moth.util.Access;
import pt.moth.util.IAccess;
import pt.moth.util.SpecialAccess;

import soot.toolkits.scalar.AbstractFlowSet;
import soot.toolkits.scalar.FlowSet;

/**
 * This FlowSet/Lattice keeps the set of accesses (reads ou writes). An acess
 * is a triple (r|w, var, boolean)
 * 
 * @author Vasco Pessanha 
 */
public class Lattice extends AbstractFlowSet {
	
	//This is a special variable that allows us to know if a given method was called by an atomic method or not..
	public final static Lattice ATOMIC_LATTICE = new Lattice(new SpecialAccess());
	
	private Set<IAccess> variables = new HashSet<IAccess>();
	
	private Lattice(IAccess a) {
		this.variables.add(a);
	}
	
	public Lattice() {}
	
	/**
	 * This method is explained in the paper PADTAD2011
	 * Algorithm adding (alfa,v,gama):
	 *  1) V\{(alfa,v,gama)} U {(alfa,v,miu)}	if	(alfa,v,gama)€ V and alfa = r and gama |= miu
	 *  2) V\{(r,v,gama)} U {(r,v,true)}		if	(r,v,gama) 	 €V and alfa = w
	 *  	U {(alfa,v,true)}
	 *  3) V U {(alfa,v,gama)}					otherwise
	 */
	public void add(Object paramObject) {
		/*//not an access
		if(!(paramObject instanceof IAccess)){
			return;
		}
		//special access
		if(paramObject instanceof SpecialAccess){
			variables.add((SpecialAccess)paramObject);
			return;
		}
		Access a = ((Access)paramObject).copy();
		Access aux = new Access(false, a.getVariable(), !a.closedBullet());
		
		//1) inserting (r,v,gama) and (r,v,!gama) is already in the view...
		if(a.isRead() && contains(aux)){
			variables.remove(aux);
			variables.add(a.copy());
		}
		// 2) inserting (w,v,?) and (r,v,?) is already in the view...
		else if(a.isWrite() && containsWithoutBullet(aux)){
			removeIgnoreType(aux);
			variables.add(new Access(true, a.getVariable().getJinstance(), true));
			variables.add(new Access(false, a.getVariable().getJinstance(), true));
		}
		// 3) Otherwise
		else{
			variables.add(a.copy());
		}*/
		if(paramObject instanceof IAccess){
			variables.add((IAccess)paramObject);
		}
	}

	@Override
	public AbstractFlowSet clone() {
		Lattice lat = new Lattice();
		for(IAccess a : variables){
			lat.add(a.copy());
		}
		return lat;
	}

	@Override
	public boolean contains(Object p) {
		for(IAccess a: variables)
			if(a.equals(p))
				return true;
		
		return false;
	}
//	public boolean containsWithoutBullet(Object paramObject){
//		for(IAccess a: variables)
//			if(a instanceof Access && ((Access)a).equalsWithoutBullet(paramObject))
//				return true;
//		
//		return false;
//	}
	public boolean containsIgnoreType(IAccess p) {
		for(IAccess a: variables)
			if(p.equalsIgnoreType(a))
				return true;
		
		return false;
	}

	@Override
	public boolean isEmpty() {
		return (variables.size() == 0);
	}

	@Override
	public void remove(Object paramObject) {
		if(paramObject instanceof IAccess)
			variables.remove(paramObject);
	}

	//Removes all accesses to this accesses' variable
	//This method does not worry with the type of access
	public void removeIgnoreType(Access param) {
		List<IAccess> toRemove = new ArrayList<IAccess>();
		//Store every access to the same variable
		for(IAccess a : variables){
			if(a instanceof Access && ((Access) a).getVariable().equals(param.getVariable())){
				toRemove.add(a);
			}
		}
		//Remove those accesses
		for(IAccess a : toRemove){
			variables.remove(a);
		}
	}
	@Override
	public int size() {
		return variables.size();
	}

	@Override
	public List<IAccess> toList() {
		List<IAccess> listToReturn = new ArrayList<IAccess>(variables.size());
		for(IAccess a: variables)
			listToReturn.add(a);
		return listToReturn;
	}
	public String toString(){
		if(variables.size() == 0)
			return "{}";
		String result = "{";
		for(IAccess a : variables){
			result+= a.toString()+", ";
		}
		return result.substring(0, result.length()-2)+"}";
	}
	@Override
	public void clear() {
		variables.clear();
	}

	@Override
	public void copy(FlowSet dest) {
		dest.clear();
		for(IAccess a : variables){
			dest.add(a.copy());
		}
	}

	@Override
	public Iterator<IAccess> iterator() {
		return variables.iterator();
	}

	//Union is the "mergep" function of PADTAD2011
	//"aux" is the result of a method call
	@Override
	public void union(FlowSet aux, FlowSet out) {
		this.copy(out);
		/*//First we add the Special Access
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof SpecialAccess){
				out.add(a.copy());
				break;
			}
		}
		//Then we add the write acesses
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof Access && ((Access)a).isWrite()){
				out.add(a.copy());
			}
		}
		//Then the read accesses
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof Access && ((Access)a).isRead()){
				out.add(a.copy());
			}
		}*/
		for(IAccess a : ((Lattice) aux).variables){
			out.add(a.copy());
		}
	}
	
	//Union is the "mergep" function of PADTAD2011
	@Override
	public void union(FlowSet aux) {
		/*//First we add the Special Access
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof SpecialAccess){
				add(a.copy());
				break;
			}
		}
		//Then we add the write acesses
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof Access && ((Access)a).isWrite()){
				add(a.copy());
			}
		}
		//Then the read accesses
		for(IAccess a : ((Lattice) aux).variables){
			if(a instanceof Access && ((Access)a).isRead()){
				add(a.copy());
			}
		}*/
		for(IAccess a : ((Lattice)aux).toList()){
			add(a.copy());
		}
	}

//	public void unionReturn(Lattice aux, Lattice dest) {
//		dest.clear();
//		aux.copy(dest);
//		for(IAccess a : variables)
//			dest.addReturn(a);
//	}
//	
//	// This function adds an Access being pessimistic.
//	// Therefore, we will always prefer an opened access rather than a closed bullets 
//	private void addReturn(IAccess obj) {
//		if(obj instanceof SpecialAccess){
//			variables.add(obj);
//		}
//		else if(obj instanceof Access){
//			Access a = (Access)obj;
//			//aux is equals to "a" but it is not closed
//			Access aux = new Access(a.isWrite(), a.getVariable().getJinstance(),!a.closedBullet());
//			if(!a.closedBullet()){
//				//take off the closed accesses to that variable
//				variables.remove(aux);
//				//put this access
//				variables.add(a);
//			}
//			else if(!variables.contains(aux)){
//				//we are trying to add a closed access and there is no acccess to that var
//				variables.add(a);
//			}
//			else{
//				//we are trying to add a closed access and we already have a opened one
//			}
//		}
//	}
	
	@Override
	public Lattice emptySet() {
		return new Lattice();
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result
				+ ((variables == null) ? 0 : variables.hashCode());
		return result;
	}

	public boolean equalsIgnoreCase(Lattice l){
		return (this.containsEveryAccessOf(l) && l.containsEveryAccessOf(this));
	}
	private boolean containsEveryAccessOf(Lattice lattice) {
		for(IAccess a : lattice.variables){
			if(!containsIgnoreType(a)){
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!(obj instanceof Lattice))
			return false;
		
		Lattice other = (Lattice) obj;
		if (variables.size() != other.size())
			return false;
		
		for(IAccess a : other.toList()){
			if(!variables.contains(a)){
				return false;
			}
		}
		return true;
	}
	@Override
	public void add(Object obj, FlowSet dest) {
		System.err.println("(Lattice)ERROR! - method add(Object obj, FlowSet dest)");
		assert(false);
	}

	@Override
	public void difference(FlowSet arg0, FlowSet arg1) {
		System.err.println("(Lattice)ERROR! - method difference(FlowSet arg0, FlowSet arg1)");
		assert(false);
	}

	@Override
	public void difference(FlowSet other) {
		System.err.println("(Lattice)ERROR! - method difference(FlowSet other)");
		assert(false);
	}

	@Override
	public void intersection(FlowSet arg0, FlowSet arg1) {
		System.err.println("(Lattice)ERROR! - method intersection(FlowSet arg0, FlowSet arg1)");
		assert(false);
	}

	@Override
	public void intersection(FlowSet other) {
		System.err.println("(Lattice)ERROR! - method intersection(FlowSet other)");
		assert(false);
	}

	@Override
	public void remove(Object obj, FlowSet dest) {
		System.err.println("(Lattice)ERROR! - method remove(Object obj, FlowSet dest)");
		assert(false);
	}
}
